/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <glob.h>

#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "gcx.h"
#include "catalogs.h"
#include "gui.h"
#include "obsdata.h"
#include "sourcesdraw.h"
#include "params.h"
#include "recipy.h"
#include "filegui.h"

#define FILE_FILTER_SZ 255

struct file_action {
	GtkWidget *filesel;
	GtkWidget *window;
	void *data;
	int arg1;
	char file_filter[FILE_FILTER_SZ+1];
};

gint destroy_filesel(gpointer data)
{
	gtk_object_set_data(GTK_OBJECT(data), "file_selector", NULL);
	return TRUE;
}

/*
static void filter_files (GtkCList *clist,
			  gint row,
			  gint column,
			  GdkEventButton *event,
			  gpointer user_data)
{
	struct file_action *fa = user_data;
	d3_printf("filter files\n");
	if (fa->file_filter[0] != 0)
		gtk_file_selection_complete(GTK_FILE_SELECTION(fa->filesel), fa->file_filter);
}
*/

GtkWidget * create_file_selection(char *title, GtkSignalFunc file_action,
				  struct file_action *action) 
{
	GtkWidget *file_selector;
	file_selector = gtk_file_selection_new(title);
	gtk_window_set_position (GTK_WINDOW (file_selector), GTK_WIN_POS_CENTER);
   
	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
			    "clicked", GTK_SIGNAL_FUNC (file_action), action);

//	gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->dir_list),
//			    "select-row", GTK_SIGNAL_FUNC (filter_files), action);

	gtk_signal_connect_object (GTK_OBJECT (file_selector), "delete_event",
                        GTK_SIGNAL_FUNC (destroy_filesel), (gpointer) action->window);
   			   
	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->ok_button),
				   "clicked", GTK_SIGNAL_FUNC (destroy_filesel),
				   (gpointer) action->window);

	gtk_signal_connect_object (GTK_OBJECT (GTK_FILE_SELECTION(file_selector)->cancel_button),
				   "clicked", GTK_SIGNAL_FUNC (destroy_filesel),
				   (gpointer) action->window);
      
	gtk_object_set_data_full(GTK_OBJECT(file_selector), "file_action", action, 
			    GTK_SIGNAL_FUNC (free));

	gtk_object_set_data_full(GTK_OBJECT(action->window), "file_selector", file_selector, 
			    GTK_SIGNAL_FUNC (gtk_widget_destroy));

	gtk_widget_show (file_selector);
	return file_selector;
}

/* return a malloced string that contains the supplied 
 * filename with ".ext" added (if it didn;t already end in '.ext')
 */
char *add_extension(char *fn, char *ext)
{
	int ns, i;
	char *fne;
	i = strlen(fn);
	ns = strlen(ext) + i + 1;

	fne = calloc(1, ns+1);
	if (fne == NULL) {
		err_printf("add_extension: alloc error\n");
		return NULL;
	}
	while (i > 0) {
		if (fn[i] == '.' || fn[i] == '/')
			break;
		i--;
	}
	if (fn[i] == '.' && !strcasecmp(fn+i+1, ext)) {
		strcpy(fne, fn);
		return fne;
	}
	strcpy(fne, fn);
	i = strlen(fn);
	fne[i] = '.';
	strcpy(fne+i+1, ext);
	return fne;
}


/* return a malloced string that contains the supplied 
 * filename with the extention (if any) replaced by ext
 */
char *force_extension(char *fn, char *ext)
{
	int ns, i;
	char *fne;
	i = strlen(fn);
	ns = strlen(ext) + i + 5;

	fne = calloc(1, ns+1);
	if (fne == NULL) {
		err_printf("force_extension: alloc error\n");
		return NULL;
	}
	while (i > 0) {
		if (fn[i] == '.' || fn[i] == '/')
			break;
		i--;
	}
	if (fn[i] == '.') {
		memcpy(fne, fn, i+1);
		strcpy(fne+i+1, ext);
		return fne;
	}
	strcpy(fne, fn);
	i = strlen(fn);
	fne[i] = '.';
	strcpy(fne+i+1, ext);
	return fne;
}


/* file action functions */

/* create a pnm file from the channel-mapped image 
 * if fa->arg1 = 1, we create a 16-bit pnm file (still TBD)
 */
static void export_pnm(GtkFileSelection *selector, gpointer user_data) 
{
	char *fn;
	struct file_action *fa = user_data;
	fn = gtk_file_selection_get_filename (GTK_FILE_SELECTION(fa->filesel));
	fn = add_extension(fn, "pnm");
	d3_printf("Saving pnm file: %s\n", fn);
	channel_to_pnm_file(fa->data, fa->window, fn);
	free(fn);
}

/* open a fits file and display it; reset the frame list to contain just this 
 * file */
static void open_fits(GtkFileSelection *selector, gpointer user_data) 
{
	char *fn, *fnd;
	struct ccd_frame *fr;
	struct file_action *fa = user_data;
	fn = gtk_file_selection_get_filename (GTK_FILE_SELECTION(fa->filesel));
	d3_printf("opening fits file: %s\n", fn);
	fr = read_gz_fits_file(fn, P_STR(FILE_UNCOMPRESS));
	if (fr == NULL) {
		statusbar2_message(fa->window, " Error trying to open fits file", "file", 5000);
		return;
	}
	rescan_fits_wcs(fr, &fr->fim);
	rescan_fits_exp(fr, &fr->exp);
	frame_to_channel(fr, fa->window, "i_channel");
	release_frame(fr);
	gtk_object_set_data_full(GTK_OBJECT(fa->window), "last_open_fits", fnd = strdup(fn), 
			    GTK_SIGNAL_FUNC (free));
	d3_printf("lastopen_fits set to %s\n", fnd);
}

#define MAX_RCP_STARS 16384
/* load stars from a rcp file into the given windows's gsl */
int load_rcp_to_window(gpointer window, char *name)
{
	struct cat_star *csl[MAX_RCP_STARS];

	int ret, p=0;
	FILE *rfn;
	char *cmd;

	if (name == NULL)
		return -1;
//	d3_printf("read_rcp: looking for %s\n", name);

	if (file_is_zipped(name)) {
		asprintf(&cmd, "%s %s ", P_STR(FILE_UNCOMPRESS), name);
		rfn = popen(cmd, "r");
		free(cmd);
		p=1;
	} else {
		rfn = fopen(name, "r");
	}
	if (rfn == NULL) {
		err_printf("read_rcp: cannot open file %s\n", name);
		return -1;
	}

	ret = read_rcp(csl, rfn, MAX_RCP_STARS, NULL, NULL);

	d3_printf("read_rcp: got %d\n", ret);

	if (p) {
		pclose(rfn);
	} else {
		fclose(rfn);
	}

	if (ret >= 0) {
		remove_stars(window, TYPE_MASK_PHOT, 0);
		add_cat_stars_to_window(window, csl, ret);
	}
	return ret;
}

#define MAX_GSC2_STARS 16384
/* load stars from a rcp file into the given windows's gsl */
int load_gsc2_to_window(gpointer window, char *name)
{
	struct cat_star *csl[MAX_GSC2_STARS];
	int ret, p=0;
	FILE *rfn;
	char *cmd;

	if (name == NULL)
		return -1;
	d3_printf("read_gsc2: looking for %s\n", name);

	if (file_is_zipped(name)) {
		asprintf(&cmd, "%s %s ", P_STR(FILE_UNCOMPRESS), name);
		rfn = popen(cmd, "r");
		free(cmd);
		p=1;
	} else {
		rfn = fopen(name, "r");
	}
	if (rfn == NULL) {
		err_printf("read_gsc2: cannot open file %s\n", name);
		return -1;
	}

	ret = read_gsc2(csl, rfn, MAX_GSC2_STARS, NULL, NULL);

	d3_printf("read_gsc2: got %d\n", ret);

	if (p) {
		pclose(rfn);
	} else {
		fclose(rfn);
	}

	if (ret >= 0) {
//		remove_stars(window, TYPE_MASK_PHOT, 0);
		add_cat_stars_to_window(window, csl, ret);
	}
	return ret;
}


/* open a rcp file and load stars; */
static void open_rcp(GtkFileSelection *selector, gpointer user_data) 
{
	char *fn, *fnd;
	struct file_action *fa = user_data;

	fn = gtk_file_selection_get_filename (GTK_FILE_SELECTION(fa->filesel));
	g_return_if_fail(fn != NULL);

	load_rcp_to_window(fa->window, fn);

	gtk_object_set_data_full(GTK_OBJECT(fa->window), "last_open_rcp", fnd = strdup(fn), 
			    GTK_SIGNAL_FUNC (free));
	d3_printf("lastopen_rcp set to %s\n", fnd);
	gtk_widget_queue_draw(fa->window);
}

/* open a gsc2 file and load stars; star that have v photometry are
 * marked as object, while the other are made field stars */
static void open_gsc2(GtkFileSelection *selector, gpointer user_data) 
{
	char *fn, *fnd;
	struct file_action *fa = user_data;

	fn = gtk_file_selection_get_filename (GTK_FILE_SELECTION(fa->filesel));
	g_return_if_fail(fn != NULL);

	load_gsc2_to_window(fa->window, fn);

	gtk_object_set_data_full(GTK_OBJECT(fa->window), "last_open_etc", fnd = strdup(fn), 
			    GTK_SIGNAL_FUNC (free));
	d3_printf("lastopen_etc set to %s\n", fnd);
	gtk_widget_queue_draw(fa->window);
}


/* set the entry with the file name */
static void set_entry(GtkFileSelection *selector, gpointer user_data) 
{
	char *fn, *fnd;
	struct file_action *fa = user_data;

	fn = gtk_file_selection_get_filename (GTK_FILE_SELECTION(fa->filesel));
	g_return_if_fail(fn != NULL);

	gtk_entry_set_text(GTK_ENTRY(fa->data), fn);

	gtk_object_set_data_full(GTK_OBJECT(fa->window), "last_open_etc", fnd = strdup(fn), 
			    GTK_SIGNAL_FUNC (free));
	d3_printf("lastopen_etc set to %s\n", fnd);
}




/* entry points */

/* make the file selector set the text in entry */
void file_select_to_entry(gpointer data, GtkWidget *entry, char *title, char *filter)
{
	GtkWidget *window = data;
	GtkWidget *filesel;
	struct file_action *fa;

	char *lastopen;
	lastopen = gtk_object_get_data(GTK_OBJECT(window), "last_open_etc");
	fa = calloc(1, sizeof(struct file_action));
	fa->window = window;
	fa->data = entry;
	if (title == NULL)
		filesel = create_file_selection("Select file", set_entry, fa);
	else
		filesel = create_file_selection(title, set_entry, fa);
	strncpy(fa->file_filter, filter, FILE_FILTER_SZ);
	fa->filesel = filesel;
	if (lastopen != NULL) {
		gtk_file_selection_set_filename (GTK_FILE_SELECTION(filesel), 
						 lastopen);
	}
}

void file_popup_cb(gpointer data, guint action, GtkWidget *menu_item)
{
	GtkWidget *window = data;
	GtkWidget *filesel;
	struct image_channel *channel;
	struct file_action *fa;
	char *fn;
	char *lastopen;
	switch(action) {
	case FILE_OPEN:
		lastopen = gtk_object_get_data(GTK_OBJECT(window), "last_open_fits");
		fa = calloc(1, sizeof(struct file_action));
		fa->window = window;
		filesel = create_file_selection("Select fits file to open", open_fits, fa);
		fa->filesel = filesel;
		strncpy(fa->file_filter, "*.fits*", FILE_FILTER_SZ);
		if (lastopen != NULL) {
			gtk_file_selection_set_filename (GTK_FILE_SELECTION(filesel), 
							 lastopen);
		}
		break;
	case FILE_OPEN_RCP:
		lastopen = gtk_object_get_data(GTK_OBJECT(window), "last_open_rcp");
		fa = calloc(1, sizeof(struct file_action));
		fa->window = window;
		filesel = create_file_selection("Select fits file to open", open_rcp, fa);
		fa->filesel = filesel;
		if (lastopen != NULL) {
			gtk_file_selection_set_filename (GTK_FILE_SELECTION(filesel), 
							 lastopen);
		}
//		gtk_file_selection_complete(GTK_FILE_SELECTION(filesel), "*.fits*");
		break;
	case FILE_LOAD_GSC2:
		lastopen = gtk_object_get_data(GTK_OBJECT(window), "last_open_etc");
		fa = calloc(1, sizeof(struct file_action));
		fa->window = window;
		filesel = create_file_selection("Select GSC-2 data file to open", open_gsc2, fa);
		fa->filesel = filesel;
		if (lastopen != NULL) {
			gtk_file_selection_set_filename (GTK_FILE_SELECTION(filesel), 
							 lastopen);
		}
//		gtk_file_selection_complete(GTK_FILE_SELECTION(filesel), "*.fits*");
		break;
	case FILE_EXPORT_PNM8:
	case FILE_EXPORT_PNM16:
		channel = gtk_object_get_data(GTK_OBJECT(window), "i_channel");
		if (channel == NULL || channel->fr == NULL)
			return ;
		fn = force_extension(channel->fr->name, "pnm");
		fa = calloc(1, sizeof(struct file_action));
		fa->window = window;
		fa->data = channel;
//		fa->arg1 = (action == FILE_EXPORT_PNM16);
		filesel = create_file_selection("Select pnm file name", export_pnm, fa);
		fa->filesel = filesel;
		gtk_file_selection_set_filename (GTK_FILE_SELECTION(filesel), fn);
		free(fn);
		break;
	default:
		err_printf("unknown action %d in file_popup_cb\n", action);
	}

}

/* return the first filename matching pattern that is found in path,
   or null if it couldn't be found */
char *find_file_in_path(char *pattern, char *path)
{
	char *dir;
	char *fn = NULL;
	char buf[1024];
	char pathc[1024];
	glob_t gl;
	int ret;

	strncpy(pathc, path, 1023);
	pathc[1023] = 0;
	dir = strtok(pathc, ":");
	while (dir != NULL) {
		snprintf(buf, 1024, "%s/%s", dir, pattern);
		d3_printf("looking for %s\n", buf);
		gl.gl_offs = 0;
		gl.gl_pathv = NULL;
		gl.gl_pathc = 0;
		ret = glob(buf, GLOB_TILDE, NULL, &gl);
//		d3_printf("glob returns %d\n", ret);
		if (ret == 0) {
			fn = strdup(gl.gl_pathv[0]);
//			d3_printf("found: %s\n", fn);
			globfree(&gl);
			return fn;
		}
		globfree(&gl);
		dir = strtok(NULL, ":");	
	}
	return NULL;
}

/* check if the file is zipped (ends with .gz, .z or .zip);
 * return 1 if it does */
int file_is_zipped(char *fn)
{
	int len;
	len = strlen(fn);
	if ((len > 4) && (strcasecmp(fn + len - 3, ".gz") == 0)) {
		return 1;
	}
	if ((len > 3) && (strcasecmp(fn + len - 2, ".z") == 0)) {
		return 1;
	}
	if ((len > 5) && (strcasecmp(fn + len - 4, ".zip") == 0)) {
		return 1;
	}
	return 0;
}
