/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

/* gui functions related to par tree editing */
/* hopefully, all the mess is contained in this file */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include <gtk/gtk.h>

#include "gcx.h"
#include "params.h"
#include "catalogs.h"
#include "sourcesdraw.h"

#define PAR_SIZE 150 /* size of par entry */

static void make_par_tree_label(GcxPar p, char *buf)
{
	if (PAR(p)->flags & PAR_USER) {
		sprintf(buf, "%s (changes)", PAR(p)->comment);
	} else {
		sprintf(buf, "%s", PAR(p)->comment);
	}
}

/* regenerate the tree label and update */
static void update_tree_label(GcxPar pp)
{
	GtkWidget *item;
	char text[256] = "";

	if (PAR_TYPE(pp) != PAR_TREE)
		return;
	item = PAR(pp)->item;
	make_par_tree_label(pp, text);
	gtk_label_set_text(GTK_LABEL(GTK_BIN(item)->child), text);
	gtk_widget_queue_draw(item);
}

static void par_item_update_status(GcxPar p)
{
	GtkWidget *item;
	GtkWidget *statlabel;
	item = PAR(p)->item;
/*  	savecheck = gtk_object_get_data(GTK_OBJECT(item), "savecheck"); */
	statlabel = gtk_object_get_data(GTK_OBJECT(item), "statlabel");
/*  	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(savecheck),  */
/*  				      ((PAR(p)->flags & PAR_TO_SAVE) != 0));  */
  	gtk_label_set_text(GTK_LABEL(statlabel), status_string(p)); 
	gtk_widget_queue_draw(item);
}

/* mark a par and all it's ancestors as changed */
static void par_mark_changed(GcxPar p)
{
	GcxPar pp;

	PAR(p)->flags |= (PAR_TO_SAVE | PAR_USER);
//	PAR(p)->flags &= ~(PAR_FROM_RC);

	par_item_update_status(p);
	pp = PAR(p)->parent;
	while (pp != PAR_NULL) {
		PAR(pp)->flags |= (PAR_TO_SAVE | PAR_USER);
//		PAR(pp)->flags &= ~(PAR_FROM_RC);
		update_tree_label(pp);
		pp = PAR(pp)->parent;
	}
}

/* we reach here when a parameter has possibly been edited */
/* this runs in tree callback context */
static void param_edited(GtkEditable *editable, GcxPar p)
{
	char *text;
	char buf[256];

	text = gtk_editable_get_chars(editable, 0, -1);
//	d3_printf("possibly edited %d text is %s\n", p, text);
	make_value_string(p, buf);
	if (!strcmp(buf, text)) {
//		d3_printf("value is the same\n");
		free(text);
		return;
	} 
	if (!try_update_par_value(p, text)) { // we have a successful update
		make_value_string(p, buf);
		gtk_entry_set_text(GTK_ENTRY(editable), buf);
		par_mark_changed(p);
	} else {
		d3_printf("cannot parse text, reverting to old value\n");
	}
	free(text);
}

/* update the item's value in the entry/optionmenu/checkbox
 */
static void par_item_update_value(GcxPar p)
{
	char val[256];
	GtkWidget *widget, *item;
	item = PAR(p)->item;
	if (PAR_TYPE(p) == PAR_INTEGER) {
		if ((PAR(p)->flags & PAR_FORMAT_FM) == FMT_OPTION) {
			widget = gtk_object_get_data(GTK_OBJECT(item), "optionmenu1");
			gtk_option_menu_set_history (GTK_OPTION_MENU (widget), P_INT(p));
			return;
		} else if ((PAR(p)->flags & PAR_FORMAT_FM) == FMT_BOOL) {
			widget = gtk_object_get_data(GTK_OBJECT(item), "toggleb1");
			gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(widget), P_INT(p));
			return;
		} 
	}
	widget = gtk_object_get_data(GTK_OBJECT(item), "entry");
	make_value_string(p, val);
	gtk_entry_set_text (GTK_ENTRY (widget), (val));
}


/* callbacks from tree items */
static void param_activate_cb (GtkEditable *editable, gpointer data)
{
	param_edited(editable, GPOINTER_TO_INT(data));
}

static void par_choice_activate_cb (GtkWidget *widget, gpointer data)
{
	GcxPar p, action;

	p = GPOINTER_TO_INT(data) & 0x0ffff;
	action = GPOINTER_TO_INT(data) >> 16;
//	d3_printf("choice activated on par %d, action %d\n", p, action);
	if (P_INT(p) != action) {
		P_INT(p) = action;
		par_mark_changed(p);
	}
}

static void par_bool_toggled_cb (GtkWidget *widget, gpointer data)
{
	GcxPar p;
	int val;

	p = GPOINTER_TO_INT(data);

	val = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
	if (P_INT(p) != val) {
		P_INT(p) = val;
		par_mark_changed(p);
	}
}


static gboolean focus_out_cb (GtkWidget *widget, GdkEventFocus *event, gpointer data)
{
	param_edited(GTK_EDITABLE(widget), GPOINTER_TO_INT(data));
	return FALSE;
}

/*
static void save_toggled_cb (GtkToggleButton *togglebutton, gpointer data)
{
	int state;
	state = gtk_toggle_button_get_active (togglebutton);
	if (state) {
		PAR(GPOINTER_TO_INT(data))->flags |= PAR_TO_SAVE;
	} else {
		PAR(GPOINTER_TO_INT(data))->flags &= ~PAR_TO_SAVE;
	}
//	d3_printf("save toggled on %d, state is %d\n", GPOINTER_TO_INT(data), state);
}
*/

static gboolean par_clicked_cb (GtkWidget *widget, GdkEventButton *event, gpointer root)
{
	if (event->button == 1) {
		if (event->state & GDK_SHIFT_MASK) { 
			gtk_tree_set_selection_mode (GTK_TREE(root),
						     GTK_SELECTION_MULTIPLE);
		} else {
			gtk_tree_set_selection_mode (GTK_TREE(root),
						     GTK_SELECTION_SINGLE);
		}
	}
	return TRUE;
}

GtkWidget *make_par_option_menu(GcxPar p, GtkWidget *treeitem)
{
	GtkWidget *optionmenu1;
	GtkWidget *optionmenu1_menu;
	GtkWidget *glade_menuitem;
	char ** c;
	int i = 0;

	c = PAR(p)->choices;

	optionmenu1 = gtk_option_menu_new ();
	gtk_widget_ref (optionmenu1);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "optionmenu1", optionmenu1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_set_usize (optionmenu1, PAR_SIZE, -2);
	gtk_widget_show (optionmenu1);
	optionmenu1_menu = gtk_menu_new ();
	while (*c != NULL) {
		glade_menuitem = gtk_menu_item_new_with_label (*c);
		gtk_widget_show (glade_menuitem);
		gtk_menu_append (GTK_MENU (optionmenu1_menu), glade_menuitem);
		gtk_signal_connect (GTK_OBJECT(glade_menuitem), "activate",
				    GTK_SIGNAL_FUNC(par_choice_activate_cb), 
				    GINT_TO_POINTER(p | i << 16));
		c++;
		i++;
	}
	gtk_option_menu_set_menu (GTK_OPTION_MENU (optionmenu1), optionmenu1_menu);
	gtk_option_menu_set_history (GTK_OPTION_MENU (optionmenu1), P_INT(p));
	return optionmenu1;
}

GtkWidget *make_par_boolean(GcxPar p, GtkWidget *treeitem)
{
	GtkWidget *toggleb;

	toggleb = gtk_check_button_new ();
	gtk_widget_ref (toggleb);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "toggleb1", toggleb,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_toggle_button_set_state(GTK_TOGGLE_BUTTON(toggleb), P_INT(p));
	gtk_widget_set_usize (toggleb, PAR_SIZE, -2);
	gtk_signal_connect (GTK_OBJECT(toggleb), "toggled",
			    GTK_SIGNAL_FUNC(par_bool_toggled_cb), 
			    GINT_TO_POINTER(p));
	gtk_widget_show (toggleb);
	return toggleb;
}


GtkWidget *make_par_entry(GcxPar p, GtkWidget *treeitem)
{
	GtkWidget *entry;
	char val[256];

	entry = gtk_entry_new ();
	gtk_widget_ref (entry);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "entry", entry,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (entry);
	gtk_widget_set_usize (entry, PAR_SIZE, -2);
	make_value_string(p, val);
	gtk_entry_set_text (GTK_ENTRY (entry), (val));
	gtk_signal_connect (GTK_OBJECT(entry), "activate",
			    GTK_SIGNAL_FUNC(param_activate_cb), GINT_TO_POINTER(p));
	gtk_signal_connect (GTK_OBJECT(entry), "focus-out-event",
			    GTK_SIGNAL_FUNC(focus_out_cb), GINT_TO_POINTER(p));
	return(entry);
}


/* make an tree item containing a parameter */
GtkWidget * make_ptree_leaf(GcxPar p, GtkWidget *root)
{
	GtkWidget *treeitem;
	GtkWidget *hbox1;
	GtkWidget *description;
	GtkWidget *statlabel;
/*  	GtkWidget *savecheck; */
	GtkWidget *parwidget = NULL;

//	d3_printf("making leaf item %d of type %d comment: %s\n", 
//		  p, PAR_TYPE(p), PAR(p)->comment);

	treeitem = gtk_tree_item_new ();
	gtk_signal_connect(GTK_OBJECT(treeitem), "button_press_event", 
			   GTK_SIGNAL_FUNC(par_clicked_cb), root);
	gtk_widget_show (treeitem);

	hbox1 = gtk_hbox_new (FALSE, 0);
	gtk_widget_ref (hbox1);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "hbox1", hbox1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (hbox1);
	gtk_container_add (GTK_CONTAINER (treeitem), hbox1);

	description = gtk_label_new (PAR(p)->comment);
	gtk_widget_ref (description);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "description", description,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (description);
	gtk_box_pack_end (GTK_BOX (hbox1), description, TRUE, TRUE, 0);
	gtk_misc_set_alignment (GTK_MISC (description), 0, 0.5);
	gtk_misc_set_padding (GTK_MISC (description), 5, 0);

	statlabel = gtk_label_new (status_string(p));
	gtk_widget_ref (statlabel);
	gtk_object_set_data_full (GTK_OBJECT (treeitem), "statlabel", statlabel,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (statlabel);

/* get our param widget here */
	if (PAR_TYPE(p) == PAR_INTEGER) {
		if ((PAR(p)->flags & PAR_FORMAT_FM) == FMT_OPTION) {
			parwidget = make_par_option_menu(p, treeitem);
		} else if ((PAR(p)->flags & PAR_FORMAT_FM) == FMT_BOOL) {
			parwidget = make_par_boolean(p, treeitem);
		} else {
			parwidget = make_par_entry(p, treeitem);
		}
	} else {
		parwidget = make_par_entry(p, treeitem);
	}

	gtk_box_pack_start (GTK_BOX (hbox1), parwidget, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (hbox1), statlabel, FALSE, FALSE, 0);
	gtk_misc_set_padding (GTK_MISC (statlabel), 5, 0);
	gtk_object_set_data(GTK_OBJECT (treeitem), "parameter", GINT_TO_POINTER(p));
 	PAR(p)->item = treeitem;
	return treeitem;
}

GtkWidget *make_stree_item(GcxPar p, GtkWidget *root)
{
	GtkWidget *item;
	char text[256] = "";

//	d3_printf("making tree item %d of type %d comment: %s\n", 
//		  p, PAR_TYPE(p), PAR(p)->comment);
	make_par_tree_label(p, text);
	item = gtk_tree_item_new_with_label (text);
	gtk_signal_connect(GTK_OBJECT(item), "button_press_event", 
			   GTK_SIGNAL_FUNC(par_clicked_cb), root);
	gtk_object_set_data(GTK_OBJECT (item), "parameter", GINT_TO_POINTER(p));
	gtk_widget_show (item);
 	PAR(p)->item = item;
	return item;
}

GtkWidget *make_subtree(GcxPar p, GtkWidget *item, GtkWidget *root)
{
	GtkWidget *subtree;
	GtkWidget *child;
	GtkWidget *sstree;
	GcxPar pp;

	if (PAR_TYPE(p) != PAR_TREE) {
		err_printf("not a tree param\n");
		return NULL;
	}
//	d3_printf("item is [%08x]\n", (unsigned)item);
	/* Create this item's subtree */
	subtree = gtk_tree_new();
	gtk_widget_ref (subtree);
	gtk_widget_show(subtree);
	gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);
	/* and now walk the children and add them in the subtree */
	/* nothing like recursion for tree building */
	gtk_object_set_data_full (GTK_OBJECT (item), "subtree", 
				  subtree, (GtkDestroyNotify) gtk_widget_unref);
//	d3_printf("subtree is [%08x]\n", (unsigned)subtree);
	pp = PAR(p)->child;
	while (pp != PAR_NULL) {
		if (PAR_TYPE(pp) == PAR_TREE) {
			child = make_stree_item(pp, root);
			gtk_tree_append (GTK_TREE(subtree), child);
			sstree = make_subtree(pp, child, root);
		} else {
			child = make_ptree_leaf(pp, root);
			gtk_tree_append (GTK_TREE(subtree), child);
		}
		pp = PAR(pp)->next;
	} 
	return item;
}

GtkWidget* make_par_tree(void) 
{
	GtkWidget *tree1;
	GcxPar p;
	GtkWidget *item;

	tree1 = gtk_tree_new ();
	p = PAR_FIRST;
	while (p != PAR_NULL) {
		item = make_stree_item(p, tree1);
// 		d3_printf("made first-level tree item %d [%08x]\n", p, (unsigned)item);
		gtk_tree_append (GTK_TREE(tree1), item);
		make_subtree(p, item, tree1);
		p = PAR(p)->next;
	}
	gtk_widget_show (tree1);
	return tree1;
}

void close_par_dialog( GtkWidget *widget, gpointer data )
{
	gtk_object_set_data(GTK_OBJECT(data), "params_window", NULL);
}

static void par_item_restore_default_gui(GcxPar p)
{
	if (PAR(p)->flags & (PAR_TREE)) { // recurse tree
		GcxPar pp;
		PAR(p)->flags &= ~PAR_USER;
		update_tree_label(p);
		pp = PAR(p)->child;
		while(pp != PAR_NULL) {
			par_item_restore_default_gui(pp);
			pp = PAR(pp)->next;
		}
	} else if (PAR(p)->flags & (PAR_USER)) {
			if (PAR_TYPE(p) == PAR_STRING) {
				change_par_string(p, PAR(p)->defval.s);
			} else {
				memcpy(&(PAR(p)->val), &(PAR(p)->defval), sizeof(union pval));
			}
			par_item_update_value(p);
//			d3_printf("defaulting par %d\n", p);
			PAR(p)->flags &= ~PAR_USER; 
			PAR(p)->flags &= ~PAR_TO_SAVE;
			par_item_update_status(p);
	}
}

static void update_ancestors_state_gui(GcxPar p)
{
	p = PAR(p)->parent;
	while (p != PAR_NULL && PAR_TYPE(p) == PAR_TREE) {
		int type = 0;
		PAR(p)->flags &= ~(PAR_USER | PAR_FROM_RC);
		type = or_child_type(p, 0);
		PAR(p)->flags |= (type & (PAR_USER | PAR_FROM_RC));
		update_tree_label(p);
		p = PAR(p)->parent;
	}
}

static void restore_defaults_cb( GtkWidget *widget, gpointer data )
{
	GList *i;
	GtkTree *tree=GTK_TREE(data);
  
	i = GTK_TREE_ROOT_TREE(tree)->selection;

	while (i) {
		GtkWidget *item;
		gpointer p;
		item = GTK_WIDGET (i->data);
		p = gtk_object_get_data(GTK_OBJECT(item), "parameter");
		par_item_restore_default_gui(GPOINTER_TO_INT(p));
		update_ancestors_state_gui(GPOINTER_TO_INT(p));
		i = i->next;
	}
}

static void par_save_cb( GtkWidget *widget, gpointer data )
{
	save_params_rc();
}

/* update item values for all siblings of p and their descendants
 */
static void par_update_items(GcxPar p)
{
	while (p != PAR_NULL) {
		if (PAR_TYPE(p) == PAR_TREE && PAR(p)->child != PAR_NULL) {
			par_update_items(PAR(p)->child);
		} else {
			par_item_update_value(p);
			par_item_update_status(p);
		}
		p = PAR(p)->next;
	}

}

static void par_load_cb( GtkWidget *widget, gpointer data )
{
	FILE *fp;
	fp = fopen ("gcxrc", "r");
	if (fp == NULL)
		return;
	fscan_params(fp, PAR_NULL);
	par_update_items(PAR_FIRST);
	fclose(fp);
}


GtkWidget* create_params_window (GtkWidget *window)
{
	GtkWidget *dialog1;
	GtkWidget *dialog_vbox1;
	GtkWidget *scrolledwindow1;
	GtkWidget *viewport1;
	GtkWidget *tree1;
	GtkWidget *dialog_action_area1;
	GtkWidget *table1;
	GtkWidget *par_save;
	GtkWidget *par_save_as;
	GtkWidget *par_default;
	GtkWidget *par_load_file;
	GtkWidget *par_refresh;
	GtkWidget *par_close;
	GtkTooltips *tooltips;

	tooltips = gtk_tooltips_new ();

	dialog1 = gtk_dialog_new ();
	gtk_object_set_data (GTK_OBJECT (dialog1), "dialog1", dialog1);
	gtk_window_set_title (GTK_WINDOW (dialog1), ("Gcx Options"));
//	GTK_WINDOW (dialog1)->type = GTK_WINDOW_DIALOG;
	gtk_window_set_position (GTK_WINDOW (dialog1), GTK_WIN_POS_CENTER);
//	gtk_window_set_modal (GTK_WINDOW (dialog1), TRUE);

	dialog_vbox1 = GTK_DIALOG (dialog1)->vbox;
	gtk_object_set_data (GTK_OBJECT (dialog1), "dialog_vbox1", dialog_vbox1);
	gtk_widget_show (dialog_vbox1);

	scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_ref (scrolledwindow1);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "scrolledwindow1", scrolledwindow1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (scrolledwindow1);
	gtk_box_pack_start (GTK_BOX (dialog_vbox1), scrolledwindow1, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (scrolledwindow1), 2);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolledwindow1), 
					GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);

	viewport1 = gtk_viewport_new (NULL, NULL);
	gtk_widget_ref (viewport1);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "viewport1", viewport1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (viewport1);
	gtk_container_add (GTK_CONTAINER (scrolledwindow1), viewport1);

	tree1 = make_par_tree();
//	gtk_signal_connect (GTK_OBJECT(tree1), "selection_changed",
//			    GTK_SIGNAL_FUNC(tree_sel_changed_cb), tree1);

	gtk_widget_ref (tree1);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "tree1", tree1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_tree_set_selection_mode (GTK_TREE(tree1),
				     GTK_SELECTION_MULTIPLE);
	gtk_widget_show (tree1);
	gtk_container_add (GTK_CONTAINER (viewport1), tree1);

	dialog_action_area1 = GTK_DIALOG (dialog1)->action_area;
	gtk_object_set_data (GTK_OBJECT (dialog1), "dialog_action_area1", dialog_action_area1);
	gtk_widget_show (dialog_action_area1);
	gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 3);

	table1 = gtk_table_new (2, 3, TRUE);
	gtk_widget_ref (table1);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "table1", table1,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (table1);
	gtk_box_pack_start (GTK_BOX (dialog_action_area1), table1, TRUE, TRUE, 0);
	gtk_table_set_row_spacings (GTK_TABLE (table1), 3);
	gtk_table_set_col_spacings (GTK_TABLE (table1), 3);

	par_save = gtk_button_new_with_label (("Save"));
	gtk_widget_ref (par_save);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_save", par_save,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_signal_connect (GTK_OBJECT (par_save), "clicked",
			    GTK_SIGNAL_FUNC (par_save_cb),
			    tree1);

	gtk_widget_show (par_save);
	gtk_table_attach (GTK_TABLE (table1), par_save, 0, 1, 0, 1,
			  (GtkAttachOptions) (GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_save, ("Save to Default Options File"), NULL);

	par_save_as = gtk_button_new_with_label (("Save As..."));
	gtk_widget_ref (par_save_as);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_save_as", par_save_as,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (par_save_as);
	gtk_table_attach (GTK_TABLE (table1), par_save_as, 0, 1, 1, 2,
			  (GtkAttachOptions) (GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_save_as, ("Save To a Options File"), NULL);
	gtk_widget_set_sensitive(par_save_as, 0);

	par_default = gtk_button_new_with_label (("Default"));
	gtk_widget_ref (par_default);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_default", par_default,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (par_default);
	gtk_table_attach (GTK_TABLE (table1), par_default, 2, 3, 0, 1,
			  (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_default, 
			      ("Restore Default Value of Selected Items"), NULL);
	gtk_signal_connect (GTK_OBJECT (par_default), "clicked",
			    GTK_SIGNAL_FUNC (restore_defaults_cb),
			    tree1);

	par_load_file = gtk_button_new_with_label (("Load From..."));
	gtk_widget_ref (par_load_file);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_load_file", par_load_file,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (par_load_file);
	gtk_table_attach (GTK_TABLE (table1), par_load_file, 1, 2, 0, 1,
			  (GtkAttachOptions) (GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_load_file, ("Load From a Options File"), NULL);
	gtk_signal_connect (GTK_OBJECT (par_load_file), "clicked",
			    GTK_SIGNAL_FUNC (par_load_cb),
			    tree1);
	gtk_widget_set_sensitive(par_load_file, 0);

	par_refresh = gtk_button_new_with_label (("Refresh"));
	gtk_widget_ref (par_refresh);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_refresh", par_refresh,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (par_refresh);
	gtk_table_attach (GTK_TABLE (table1), par_refresh, 1, 2, 1, 2,
			  (GtkAttachOptions) (GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_refresh, ("Refresh Values of All Options"), NULL);
	gtk_widget_set_sensitive(par_refresh, 0);

	par_close = gtk_button_new_with_label (("Close"));
	gtk_widget_ref (par_close);
	gtk_object_set_data_full (GTK_OBJECT (dialog1), "par_close", par_close,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (par_close);
	gtk_table_attach (GTK_TABLE (table1), par_close, 2, 3, 1, 2,
			  (GtkAttachOptions) (GTK_FILL),
			  (GtkAttachOptions) (0), 0, 0);
	gtk_tooltips_set_tip (tooltips, par_close, ("Close Dialog"), NULL);
	gtk_signal_connect (GTK_OBJECT (par_close), "clicked",
			    GTK_SIGNAL_FUNC (close_par_dialog),
			    window);
	gtk_object_set_data (GTK_OBJECT (dialog1), "tooltips", tooltips);
	gtk_window_set_default_size(GTK_WINDOW(dialog1), 500, 400); 
	return dialog1;
}

