/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// tele.c: telescope control functions
// $Revision: 1.1 $
// $Date: 2003/12/01 00:19:37 $

/* a very unthreadable lx200 interface; should be treated in the same way as 
 * the cameras, and piped through tcp, but it's a start */

#include <stdio.h>
#include <string.h>
#include <sys/io.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/resource.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <sched.h>
#include <math.h>
#include <errno.h>

#include "gcx.h"
#include "params.h"

#define JD2000 2451545.0
//#define	degrad(x)	((x)*PI/180.)
//#define	raddeg(x)	((x)*180./PI)
#define	hrdeg(x)	((x)*15.)
#define	deghr(x)	((x)/15.)
#define	hrrad(x)	degrad(hrdeg(x))
#define	radhr(x)	deghr(raddeg(x))

static double lx_epoch = 2000.0;
static double last_ra = 0.0;
static double last_dec = 0.0;
static double target_ra = 0.0;
static double target_dec = 0.0;
static double last_epoch = 2000.0;

#define SLEW_FINISH_TIME 5000 /* we wait 5 seconds from the time the scope coordinates
			       * have reached their target before we declare the slew over
			       */
#define SLEW_TIMEOUT 120000 /* max duration of a slew: 2 minutes */
#define LX_SERIAL_TIMEOUT 200000 /* this one is in us */
#define SLEW_END_TOL (1) 

struct timeval slew_start_time;
struct timeval slew_end_time;

// general interval timer functions
static void update_timer(struct timeval *tv_old)
{				/* init timer */
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = 0;
	gettimeofday(&tv, &tz);
	tv_old->tv_usec = tv.tv_usec;	/* update state variable */
	tv_old->tv_sec = tv.tv_sec;
}

static unsigned get_timer_delta(struct timeval *tv_old)
{				/* read the exposure timer
				   in ms from the last call to update_timer */
	unsigned int deltat;
	struct timeval tv;
	struct timezone tz;

	tz.tz_minuteswest = 0;
	gettimeofday(&tv, &tz);
	deltat = (tv.tv_sec - tv_old->tv_sec) * 1000 +
	    (tv.tv_usec - tv_old->tv_usec) / 1000;
	return (deltat);
}

// return the ms difference between t1 and t2 (positive if t1 is earlier than t2)
/*
static int time_diff(struct timeval *t1, struct timeval *t2)
{
	int deltat;
	deltat = t2->tv_sec - t1->tv_sec;
	deltat *= 1000;
	deltat += (t2->tv_usec - t2->tv_usec) / 1000;
	return deltat;
}
*/

void set_lx_epoch(double epo)
{
	lx_epoch = epo;
}

struct termios tio;
int ser;

#define MAX_RETRY 10000

int sync_time(void)
{
	int fd;
	char rbuf[256];
	int rp;
	int i, ret;
	int time;

	fd = open(P_STR(FILE_GPS_SERIAL), O_RDWR, O_NONBLOCK);
	tcgetattr(fd, &tio);
	cfsetospeed(&tio, B4800);
	cfsetispeed(&tio, B4800);
	cfmakeraw(&tio);
	tcsetattr(fd, TCSANOW, &tio);

	rp = 0;
	rbuf[255] = 0;
	for (i=0; i<MAX_RETRY; i++) {
		if (read(fd, rbuf, 1) == 1) {
//			d3_printf("Got: %c\n", *(rbuf));
			if (*(rbuf) == '\n') {//end of line
				rp = 0;
				break;
			}
			rp ++;
			if (rp > 250) {
//				err_printf("sync_time: line too long\n");
				close(fd);
				return -1;
			}
		}
	}
	rp = 0;
	for (i=0; i<MAX_RETRY; i++) {
		if (read(fd, rbuf+rp, 1) == 1) {
//			d3_printf("2l:Got: %c (%d)\n", *(rbuf+rp), *(rbuf+rp));
			if (*(rbuf+rp) == '\n') {//end of line
				rp = 0;
//				err_printf("bad line :%s\n");
				close(fd);
				return -1;
			}
			rp ++;
			if (rp > 250) {
				err_printf("sync_time: line too long\n");
				close(fd);
				return -1;
			}
			if (rp > 14 && (ret = sscanf(rbuf, "$GPRMC,%d,", &time)) == 1) {
				rbuf[15] = 0;
				d3_printf("timeline: %s\n", rbuf);
				close(fd);
				return (time);
			}
		}
	}	
	err_printf("Timeout in receive\n");
	close(fd);
	return -1;
}

void open_serial(void)
{
	ser = open(P_STR(FILE_SCOPE_SERIAL), O_RDWR);
	if (ser <= 0) {
		err_printf("serial open error\n");
	}
//	d3_printf("ser opened: %d\n", ser);
	tcgetattr(ser, &tio);
	cfsetospeed(&tio, B9600);
	cfsetispeed(&tio, B9600);
	cfmakeraw(&tio);
	tcsetattr(ser, TCSANOW, &tio);
}

void close_serial(void)
{
//	d3_printf("ser closing (%d)\n", ser);
	close(ser);
}



void sersend(char *buf)
{
//	open_serial();
//	d3_printf(">>> %s \n", buf);
	write(ser, buf, strlen(buf));
//	close_serial();
}


//dx is the move in E direction (neg for W), dy the move in N (neg for S)
//value is in ms

// minimum value of guide step (ms, must be >5)
#define MINGUIDE 5

void guide_move(double dx, double dy)
{
	struct sched_param sp;

	setpriority(PRIO_PROCESS, 0, 20);
//	printf("Guide move: %.1f %.1f\n", dx, dy);
	usleep(10000);
	open_serial();
	if (abs(dx) > MINGUIDE || abs(dy) > MINGUIDE) { 
		sp.sched_priority = 10;
		sched_setscheduler(0,SCHED_FIFO,&sp);

		sersend(":RG#");

		if (abs(dx) > MINGUIDE) {
			if (dx > 0) {	
				sersend(":Me#");
				usleep((int)(floor(1000.0 * dx)) - 4000);
			} else {
				sersend(":Mw#");
				usleep((int)(floor(1000.0 * -dx)) - 4000);
			}
			if (dx > 0) {	
				sersend(":Qe#");
			} else {
				sersend(":Qw#");
			}
		}
		if (abs(dy) > MINGUIDE) {
			if (dy > 0) {	
				sersend(":Mn#");
				usleep((int)(floor(1000.0 * dy)) - 4000);
			} else {
				sersend(":Ms#");
				usleep((int)(floor(1000.0 * -dy)) - 4000);
			}
			if (dy > 0) {	
				sersend(":Qn#");
			} else {
				sersend(":Qs#");
			}
		}
		sp.sched_priority = 0;
		sched_setscheduler(0,SCHED_OTHER,&sp);
		setpriority(PRIO_PROCESS, 0, 20);
	}
	close_serial();
}

// transform degrees into lx200 declination format
static void dec_to_dms(char *lb, double deg)
{
	double d1, d2;
	if (deg >= 0.0) {
		deg += 1.0 / (120.0); //add half a minute for correct rounding
		d1 = floor (deg);
		deg = (deg - d1) * 60.0;
		d2 = floor (deg);
		deg = ((deg - d2) * 60.0);
		sprintf(lb, "+%02d:%02d", (int)d1,  
			(int)d2);
//		sprintf(lb, "-%02d%c%02d'%02d", (int)d1, 223,
//			(int)d2, (int)floor(deg));
	} else {
		deg = -deg;
		deg += 1.0 / (120.0); //add half a minute for correct rounding
		d1 = floor (deg);
		deg = (deg - d1) * 60.0;
		d2 = floor (deg);
		deg = ((deg - d2) * 60.0);
		sprintf(lb, "-%02d:%02d", (int)d1,  
			(int)d2);
//		sprintf(lb, "-%02d%c%02d'%02d", (int)d1, 223,
//			(int)d2, (int)floor(deg));
	}
}

// transform degrees into lx200 ra format
static void ra_to_hms(char *lb, double deg)
{
	double d1, d2;
	deg /= 15.0;
	deg += 1.0 / (60.0 * 20.0); //add half a tenth of a minute for correct rounding
	if (deg >= 0.0) {
		d1 = floor (deg);
		deg = (deg - d1) * 60.0;
		d2 = floor (deg);
		deg = ((deg - d2) * 10.0);
		sprintf(lb, "%02d:%02d.%1d", (int)d1, 
			(int)d2, (int)floor(deg));
	} else {
		err_printf("bad RA (< 0), assuming positive\n");
		deg = -deg;
		d1 = floor (deg);
		deg = (deg - d1) * 60.0;
		d2 = floor (deg);
		deg = ((deg - d2) * 10.0);
		sprintf(lb, "%02d:%02d.%1d", (int)d1, 
			(int)d2, (int)floor(deg));
	}
}

/* get the current scope position (just return the object set for now)
 * return 0 if successfull */
int lx200_get_target(double *ra, double *dec, double *epoch)
{
	*ra = last_ra;
	*dec = last_dec;
	*epoch = last_epoch;
	return 0;
}


// send the 'obs' object coordinates to the telescope
int lx200_set_object(double ra, double dec, double epoch)
{
	char lb[64];
	char out[64];
	int i, j, k, jdi;

	struct timeval tv;
	struct timezone tz;
	struct tm *t;

	if (dec > 80.0) {
		err_printf("Object too close to the pole, aborting goto\n");
		return -1;
	}

	tz.tz_minuteswest = 0;
	gettimeofday(&tv, &tz);
	t = gmtime(&(tv.tv_sec));

	last_ra = ra;
	last_dec = dec;
	last_epoch = epoch;

// do the julian date (per hsaa p107)
	i = 1900 + t->tm_year;
	j = t->tm_mon + 1;
	k = t->tm_mday;

	jdi = k - 32075 + 1461 * (i + 4800 + (j - 14) / 12) / 4
		+ 367 * (j - 2 - (j - 14) / 12 * 12) / 12
		- 3 * (( i + 4900 + (j - 14) / 12) / 100) / 4;

	precess_hiprec(epoch, lx_epoch, &ra, &dec);

	ra_to_hms(lb, ra);
	d3_printf("RA'%.2f %s", lx_epoch, lb);
	dec_to_dms(lb, dec);
	d3_printf(" DEC'%.2f %s -> scope\n", lx_epoch, lb);

	target_ra = ra;
	target_dec = dec;

	d3_printf("ra %.3f dec %.3f\n", ra, dec);

	open_serial();

	dec_to_dms(lb, dec);
	sprintf(out, ":Sd %s#", lb);
	sersend(out);

	ra_to_hms(lb, ra);
	sprintf(out, ":Sr %s#", lb);
	sersend(out);

	close_serial();
	update_timer(&slew_start_time);
	update_timer(&slew_end_time);

	precess_hiprec(lx_epoch, (jdi - JD2000) / 365.25 + 2000.0, &ra, &dec);

//	ra_to_hms(lb, ra);
//	d3_printf("RA'now     %s", lb);
//	dec_to_dms(lb, dec);
//	d3_printf(" DEC'now     %s\n", lb);
	return 0;
}

// start a slew to the selected object
void lx200_start_slew(void)
{
	open_serial();
	sersend(":MS#");
	close_serial();
}

// abort current slew
void lx200_abort_slew(void)
{
	open_serial();
	sersend(":Q#");
	close_serial();
}

// sync to current coordinates
static int lx200_sync(void)
{
	open_serial();
	sersend(":CM#");
	close_serial();
	usleep(100000);
	return 0;
}


int lx200_sync_coords(double ra, double dec, double epoch)
{
	int ret;
	if (epoch == 0.0)
		epoch = 2000.0;
	ret = lx200_set_object(ra, dec, epoch);
	if (ret)
		return ret;
	ret = lx200_sync();
	return ret;
}

/* do one select and one read from the given file */
static int timeout_read(int fd, char * resp, int resp_size, int timeout)
{
	struct timeval tmo;
	int ret;
	fd_set fds;

	tmo.tv_sec = 0;
	tmo.tv_usec = timeout;
	FD_ZERO(&fds);
	FD_SET(fd, &fds);

	ret = select(FD_SETSIZE, &fds, NULL, NULL, &tmo);
	if (ret == 0) {
		err_printf("lx_read: timeout in receive\n");
		return -1;
	}
	if (ret == 1 && errno) {
		err_printf("lx_read: select error\n");
		return -1;
	}
	ret = read(fd, resp, resp_size);
//	d3_printf("read returns %d, val=%d\n", ret, *resp);
	if (ret == 1 && errno != 0) {
		err_printf("lx_read: error in receive [%s]\n", strerror(errno));
		return -1;
	}
	return ret;
}


/* read the response from the scope (until we get an '#' or the buffer fills)
 * returns -1 if a timeout occured, or the number of charaters read (including the 
 * termimating '#'. A zero is placed at the end of the response */
static int lx200_read(int fd, char *buf, int size)
{
	int len = 0, ret;
	while (size > 1) {
		ret = timeout_read(fd, buf, 1, LX_SERIAL_TIMEOUT) == 1;
//		d3_printf("lx:read: val=%d, ret %d\n", *buf, ret);
		if (ret == 1) {
//			d3_printf("lx:read: val=%d\n", *buf);

			buf ++;
			size --;
			len ++;
			if (*(buf-1) == '#') {
				*buf = 0;
				return len;
			}
		} else {
			return -1;
		}
	}
	return len;
}


/* read the position from scope. epoch is the epoch last set (usually 2000) 
 * we assume all coordinates are precessed to it 
 * returns 0 if the read was successfull, -1 for an error */
int lx200_get_position(double *ra, double *dec, double *epoch)
{
	char buf[256];
	int ret;
	double lra, ldec;

	open_serial();
	sersend(":GR#");
	ret = lx200_read(ser, buf, 256);
	if (ret < 0) {
		err_printf("Timeout talking to scope\n");
		goto err_close;
	}
//	d3_printf("<<< %s\n", buf);
	ret = dms_to_degrees(buf, &lra);
	if (ret) {
		err_printf("Error parsing ra from scope\n");
		goto err_close;
	}
	lra *= 15.0;
	sersend(":GD#");
	ret = lx200_read(ser, buf, 256);
	if (ret < 0) {
		err_printf("Timeout talking to scope\n");
		goto err_close;
	}
//	d3_printf("<<< %s\n", buf);
	ret = dms_to_degrees(buf, &ldec);
	if (ret) {
		err_printf("Error parsing dec from scope\n");
		goto err_close;
	}
	if (ra != NULL)
		*ra = lra; 
	if (dec != NULL)
		*dec = ldec;
	if (epoch != NULL)
		*epoch = lx_epoch;
	close_serial();
	return 0;

err_close:
	close_serial();
	return -1;
}

/* check if the slew has finished. Return 0 if finished, -1 if we can;t talk to the scope 
 * or 1 if slew is in progress. In the latter case, update dra and ddec with the amount 
 * of movement still needed. We always return 0 after SLEW_TIMEOUT */
int lx200_poll_slew_status(double *dra, double *ddec)
{
	double ra, dec;
	int ret;

	if (get_timer_delta (&slew_start_time) > SLEW_TIMEOUT) {
		err_printf("slew timed out\n");
		return 0;
	}

	ret = lx200_get_position(&ra, &dec, NULL);
	if (ret != 0) 
		return -1;

//	d3_printf("+++ %.3f, %.3f\n", ra, dec);

	if (fabs(ra - target_ra) > SLEW_END_TOL ||
	    fabs(dec - target_dec) > SLEW_END_TOL) {
		update_timer(&slew_end_time);
	}
	if (dra != NULL)
		*dra = fabs(ra - target_ra);
	if (ddec != NULL)
		*ddec = fabs(dec - target_dec);

	if (get_timer_delta (&slew_end_time) > SLEW_FINISH_TIME) {
		return 0;
	}
	return 1;
}
